import os
import subprocess

__OIDReceptorName = "1.3.6.1.4.1.2312.19.1"
__OIDReceptorNameFormat = "UTF8"


def __init__():
    pass


def create_certificate(tmp_dir: str, commonName: str = "localhost"):
    def generate_cert(name, commonName):
        keyPath = os.path.join(tmp_dir, name + ".key")
        crtPath = os.path.join(tmp_dir, name + ".crt")
        subprocess.check_output(["openssl", "genrsa", "-out", keyPath, "2048"])
        subprocess.check_output(
            [
                "openssl",
                "req",
                "-x509",
                "-new",
                "-nodes",
                "-key",
                keyPath,
                "-subj",
                "/C=/ST=/L=/O=/OU=ReceptorTesting/CN=ca",
                "-sha256",
                "-out",
                crtPath,
            ]
        )
        return keyPath, crtPath

    def generate_cert_with_ca(name, caKeyPath, caCrtPath, commonName):
        keyPath = os.path.join(tmp_dir, name + ".key")
        crtPath = os.path.join(tmp_dir, name + ".crt")
        csrPath = os.path.join(tmp_dir, name + ".csa")
        extPath = os.path.join(tmp_dir, name + ".ext")

        # create x509 extension
        with open(extPath, "w") as ext:
            # DNSName to SAN
            ext.write("subjectAltName=DNS:" + commonName)
            # Receptor NodeID (otherName) to SAN
            ext.write(
                ",otherName:"
                + __OIDReceptorName
                + ";"
                + __OIDReceptorNameFormat
                + ":"
                + commonName
            )
            ext.close()
        subprocess.check_output(["openssl", "genrsa", "-out", keyPath, "2048"])

        # create cert request
        subprocess.check_output(
            [
                "openssl",
                "req",
                "-new",
                "-sha256",
                "-key",
                keyPath,
                "-subj",
                "/C=/ST=/L=/O=/OU=ReceptorTesting/CN=" + commonName,
                "-out",
                csrPath,
            ]
        )

        # sign cert request
        subprocess.check_output(
            [
                "openssl",
                "x509",
                "-req",
                "-extfile",
                extPath,
                "-in",
                csrPath,
                "-CA",
                caCrtPath,
                "-CAkey",
                caKeyPath,
                "-CAcreateserial",
                "-out",
                crtPath,
                "-sha256",
            ]
        )

        return keyPath, crtPath

    # Create a new CA
    caKeyPath, caCrtPath = generate_cert("ca", "ca")
    clientKeyPath, clientCrtPath = generate_cert_with_ca(
        "client", caKeyPath, caCrtPath, commonName
    )
    generate_cert_with_ca("server", caKeyPath, caCrtPath, commonName)

    return {
        "caKeyPath": caKeyPath,
        "caCrtPath": caCrtPath,
        "clientKeyPath": clientKeyPath,
        "clientCrtPath": clientCrtPath,
    }



import receptorctl

import pytest
import subprocess
import os
import shutil
import time
import json
import yaml
from click.testing import CliRunner

from .lib import create_certificate


@pytest.fixture(scope="session")
def base_tmp_dir():
    receptor_tmp_dir = "/tmp/receptor"
    base_tmp_dir = "/tmp/receptorctltest"

    # Clean up tmp directory and create a new one
    if os.path.exists(base_tmp_dir):
        shutil.rmtree(base_tmp_dir)
    os.mkdir(base_tmp_dir)

    yield base_tmp_dir

    # Tear-down
    # if os.path.exists(base_tmp_dir):
    #     shutil.rmtree(base_tmp_dir)

    if os.path.exists(receptor_tmp_dir):
        shutil.rmtree(receptor_tmp_dir)

    subprocess.call(["killall", "receptor"])


@pytest.fixture(scope="class")
def receptor_mesh(base_tmp_dir):
    class ReceptorMeshSetup:
        # Relative dir to the receptorctl tests
        mesh_definitions_dir = "tests/mesh-definitions"

        def __init__(self):
            # Default vars
            self.base_tmp_dir = base_tmp_dir

            # Required dependencies
            self.__check_dependencies()

        def setup(self, mesh_name: str = "mesh1", socket_file_name: str = "node1.sock"):
            self.mesh_name = mesh_name
            self.__change_config_files_dir(mesh_name)
            self.__create_tmp_dir()
            self.__create_certificates()
            self.socket_file_name = socket_file_name

            # HACK this should be a dinamic way to select a node socket
            self.default_socket_unix = "unix://" + os.path.join(
                self.get_mesh_tmp_dir(), socket_file_name
            )

        def default_receptor_controller_unix(self):
            return receptorctl.ReceptorControl(self.default_socket_unix)

        def __change_config_files_dir(self, mesh_name: str):
            self.config_files_dir = "{}/{}".format(self.mesh_definitions_dir, mesh_name)
            self.config_files = []

            # Iterate over all the files in the config_files_dir
            # and create a list of all files that end with .yaml or .yml
            for f in os.listdir(self.config_files_dir):
                if f.endswith(".yaml") or f.endswith(".yml"):
                    self.config_files.append(os.path.join(self.config_files_dir, f))

        def __create_certificates(self):
            self.certificate_files = create_certificate(
                self.get_mesh_tmp_dir(), "node1"
            )

        def get_mesh_name(self):
            return self.config_files_dir.split("/")[-1]

        def get_mesh_tmp_dir(self):
            mesh_tmp_dir = "{}/{}".format(self.base_tmp_dir, self.mesh_name)
            return mesh_tmp_dir

        def __check_dependencies(self):
            """Check if we have the required dependencies
            raise an exception if we don't
            """

            # Check if openssl binary is on the path
            try:
                subprocess.check_output(["openssl", "version"])
            except FileNotFoundError:
                raise Exception(
                    "openssl binary not found\n"
                    'Consider run "sudo dnf install openssl"'
                )

        def __create_tmp_dir(self):
            mesh_tmp_dir_path = self.get_mesh_tmp_dir()

            # Clean up tmp directory and create a new one
            if os.path.exists(mesh_tmp_dir_path):
                shutil.rmtree(mesh_tmp_dir_path)
            os.mkdir(mesh_tmp_dir_path)

    return ReceptorMeshSetup()


@pytest.fixture(scope="class")
def receptor_bin_path():
    """Returns the path to the receptor binary

        This fixture was created to make possible the use of
        multiple receptor binaries files.

        The default priority order is:
        - ../../tests/artifacts-output
        - The "receptor" available in the PATH

    Returns:
        str: Path to the receptor binary
    """

    # Check if the receptor binary is in '../../tests/artifacts-output' and returns
    # the path to the binary if it is found.
    receptor_bin_path_from_test_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "../../tests/artifacts-output/",
        "receptor",
    )
    if os.path.exists(receptor_bin_path_from_test_dir):
        return receptor_bin_path_from_test_dir

    # Check if the receptor binary is in '../../' and returns
    # the path to the binary if it is found.
    receptor_bin_path_from_test_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "../../",
        "receptor",
    )
    if os.path.exists(receptor_bin_path_from_test_dir):
        return receptor_bin_path_from_test_dir

    # Check if the receptor binary is in the path
    try:
        subprocess.check_output(["receptor", "--version"])
        return "receptor"
    except subprocess.CalledProcessError:
        raise Exception(
            "Receptor binary not found in $PATH or in '../../tests/artifacts-output'"
        )


@pytest.fixture(scope="class")
def default_socket_tcp():
    return "tcp://localhost:11112"


@pytest.fixture(scope="class")
def default_socket_file(receptor_mesh):
    return receptor_mesh.get_mesh_tmp_dir() + "/node1.sock"


@pytest.fixture(scope="class")
def default_receptor_controller_socket_file(default_socket_file):
    return receptorctl.ReceptorControl(default_socket_file)


@pytest.fixture(scope="class")
def default_receptor_controller_tcp(default_socket_tcp):
    return receptorctl.ReceptorControl(default_socket_tcp)


@pytest.fixture(scope="class")
def default_receptor_controller_tcp_tls(default_socket_tcp, certificate_files):
    rootcas = certificate_files["caCrtPath"]
    key = certificate_files["clientKeyPath"]
    cert = certificate_files["clientCrtPath"]
    insecureskipverify = True

    controller = receptorctl.ReceptorControl(
        default_socket_tcp,
        rootcas=rootcas,
        key=key,
        cert=cert,
        insecureskipverify=insecureskipverify,
    )

    return controller


@pytest.fixture(scope="class")
def receptor_nodes():
    class ReceptorNodes:
        nodes = []
        log_files = []

    return ReceptorNodes()


def receptor_nodes_kill(nodes):
    for node in nodes:
        node.kill()

    for node in nodes:
        node.wait(3)


def import_config_from_node(node):
    """Receive a node and return the config file as a dict"""
    stream = open(node.args[2], "r")
    try:
        config_unflatten = yaml.safe_load(stream)
    except yaml.YAMLError as e:
        raise e
    stream.close()

    config = {}
    for c in config_unflatten:
        config.update(c)

    return config


def receptor_mesh_wait_until_ready(nodes, receptor_controller):
    time.sleep(0.5)

    # Try up to 6 times
    tries = 0
    while True:
        status = receptor_controller.simple_command("status")
        # Check if it has three known nodes
        if len(status["KnownConnectionCosts"]) == 3:
            break
        tries += 1
        if tries > 6:
            raise Exception("Receptor Mesh did not start up")
        time.sleep(1)

    receptor_controller.close()


@pytest.fixture(scope="class")
def certificate_files(receptor_mesh):
    return receptor_mesh.certificate_files


@pytest.fixture(scope="class")
def default_receptor_controller_unix(receptor_mesh):
    return receptor_mesh.default_receptor_controller_unix()


def start_nodes(receptor_mesh, receptor_nodes, receptor_bin_path):
    for i, config_file in enumerate(receptor_mesh.config_files):
        log_file_name = (
            config_file.split("/")[-1].replace(".yaml", ".log").replace(".yml", ".log")
        )
        receptor_nodes.log_files.append(
            open(
                os.path.join(receptor_mesh.get_mesh_tmp_dir(), log_file_name),
                "w",
            )
        )
        receptor_nodes.nodes.append(
            subprocess.Popen(
                [receptor_bin_path, "-c", config_file],
                stdout=receptor_nodes.log_files[i],
                stderr=receptor_nodes.log_files[i],
            )
        )


@pytest.fixture(scope="class")
def receptor_mesh_mesh1(
    receptor_bin_path,
    receptor_nodes,
    receptor_mesh,
):
    # Set custom config files dir
    receptor_mesh.setup("mesh1")

    # Start the receptor nodes processes
    start_nodes(receptor_mesh, receptor_nodes, receptor_bin_path)

    receptor_mesh_wait_until_ready(
        receptor_nodes.nodes, receptor_mesh.default_receptor_controller_unix()
    )

    yield

    receptor_nodes_kill(receptor_nodes.nodes)


@pytest.fixture(scope="class")
def receptor_mesh_access_control(
    receptor_bin_path,
    receptor_nodes,
    receptor_mesh,
):
    # Set custom config files dir
    receptor_mesh.setup("access_control", "node2.sock")

    # Create PEM key for signed work
    key_path = os.path.join(receptor_mesh.get_mesh_tmp_dir(), "signwork_key")
    subprocess.check_output(
        [
            "ssh-keygen",
            "-b",
            "2048",
            "-t",
            "rsa",
            "-f",
            key_path,
            "-q",
            "-N",
            "",
        ]
    )

    # Start the receptor nodes processes
    start_nodes(receptor_mesh, receptor_nodes, receptor_bin_path)

    receptor_mesh_wait_until_ready(
        receptor_nodes.nodes, receptor_mesh.default_receptor_controller_unix()
    )

    yield

    receptor_nodes_kill(receptor_nodes.nodes)


@pytest.fixture(scope="function")
def receptor_control_args(receptor_mesh):
    args = {
        "--socket": f"{receptor_mesh.get_mesh_tmp_dir()}/{receptor_mesh.socket_file_name}",
        "--config": None,
        "--tls": None,
        "--rootcas": None,
        "--key": None,
        "--cert": None,
        "--insecureskipverify": None,
    }
    return args


@pytest.fixture(scope="function")
def invoke(receptor_control_args):
    def f_invoke(command, args: list = []):
        """Invoke a command and return the original result.

        Args:
            command (click command): The command to invoke.
            args (list<str>): The arguments to pass to the command.

        Returns:
            click.testing: The original result.
        """

        def parse_args_to_list(args: dict):
            """Parse the args (dict) to a list of strings."""
            arg_list = []
            for k, v in args.items():
                if v is not None:
                    arg_list.append(str(k))
                    arg_list.append(str(v))
            return arg_list

        # Since we may log errors/warnings on stderr we want to split stdout and stderr
        runner = CliRunner(mix_stderr=False)

        out = runner.invoke(
            receptorctl.cli.cli,
            parse_args_to_list(receptor_control_args) + [command.name] + args,
        )
        return out

    return f_invoke


@pytest.fixture(scope="function")
def invoke_as_json(invoke):
    def f_invoke_as_json(command, args: list = []):
        """Invoke a command and return the original result and the json output.

        Args:
            command (click command): The command to invoke.
            args (list<str>): The arguments to pass to the command.

        Returns:
            tuple<click.testing, dict>: Tuple of the original result and the json output.
        """
        result = invoke(command, ["--json"] + args)
        try:
            # JSON data should only be on stdout
            json_output = json.loads(result.stdout)
        except json.decoder.JSONDecodeError:
            pytest.fail("The command is not in json format")
        return result, json_output

    return f_invoke_as_json



# The goal is to write tests following the click documentation:
# https://click.palletsprojects.com/en/8.0.x/testing/

import pytest
import time


@pytest.fixture(scope="function")
def wait_for_workunit_state():
    def _wait_for_workunit_state(
        node_controller,
        unitid: str,
        expected_detail: str = None,
        expected_state_name: str = None,
        timeout_seconds: int = 30,
    ) -> bool:
        """Wait for a workunit to finish

        At least 'expected_detail' or 'expected_state_name' must be specified.

        Args:
            node_controller: The node controller used to create the workunit
            unitid: The unitid of the workunit to wait for
            expected_detail: The expected detail of the workunit
            expected_state_name: The expected state name of the workunit
            timeout_seconds: The number of seconds to wait before timing out

        Returns:
            True if the workunit finished, False if it timed out
        """
        if expected_detail is None and expected_state_name is None:
            raise ValueError(
                "At least 'expected_detail' or 'expected_state_name' must be specified"
            )

        remaining_time = timeout_seconds

        if expected_detail is not None:
            for _ in range(remaining_time):
                status = node_controller.simple_command("work status {}".format(unitid))
                if status["Detail"] == expected_detail:
                    return True
                else:
                    time.sleep(1)
                    remaining_time -= 1

        if remaining_time <= 0:
            return False

        if expected_state_name is not None:
            for _ in range(remaining_time):
                status = node_controller.simple_command("work status {}".format(unitid))
                if status["StateName"] == expected_state_name:
                    return True
                else:
                    time.sleep(1)
                    remaining_time -= 1
        return False

    return _wait_for_workunit_state


@pytest.fixture(scope="function")
def wait_for_work_finished(wait_for_workunit_state):
    def _wait_for_work_finished(
        node_controller, unitid: str, timeout_seconds: int = 30
    ) -> bool:
        """Wait for a workunit to finish

        Args:
            node_controller: The node controller used to create the workunit
            unitid: The unitid of the workunit to wait for
            timeout_seconds: The number of seconds to wait before timing out

        Returns:
            True if the workunit finished, False if it timed out
        """

        return wait_for_workunit_state(
            node_controller,
            unitid,
            expected_detail="exit status 0",
            expected_state_name="Succeeded",
            timeout_seconds=timeout_seconds,
        )

    return _wait_for_work_finished


@pytest.mark.usefixtures("receptor_mesh_mesh1")
class TestWorkUnit:
    def test_workunit_simple(
        self,
        invoke_as_json,
        default_receptor_controller_socket_file,
        wait_for_work_finished,
    ):
        # Spawn a long running command
        node1_controller = default_receptor_controller_socket_file

        wait_for = 5  # in seconds

        payload = "That's a long string example! And there's emoji too! 👾"
        work = node1_controller.submit_work("echo-uppercase", payload, node="node3")
        state_result = work.pop("result")
        state_unitid = work.pop("unitid")

        assert state_result == "Job Started"
        assert wait_for_work_finished(
            node1_controller, state_unitid, wait_for
        ), "Workunit timed out and never finished"

        work_result = (
            node1_controller.get_work_results(state_unitid)
            .read()
            .decode("utf-8")
            .strip()
        )

        assert payload.upper() == work_result, (
            f"Workunit did not report the expected result:\n - payload: {payload}"
            f"\n - work_result: {work_result}"
        )

        node1_controller.close()

    def test_workunit_cmd_cancel(
        self,
        invoke_as_json,
        default_receptor_controller_socket_file,
        wait_for_workunit_state,
    ):
        # Spawn a long running command
        node1_controller = default_receptor_controller_socket_file

        sleep_for = 9999  # in seconds
        wait_for = 15  # in seconds

        work = node1_controller.submit_work("sleep", str(sleep_for), node="node3")
        state_result = work.pop("result")
        state_unitid = work.pop("unitid")
        assert state_result == "Job Started"

        # HACK: Wait for the workunit to start
        # receptor should be able to cancel the workunit with this
        time.sleep(5)

        # Run and check cancel command
        cancel_output = node1_controller.simple_command(f"work cancel {state_unitid}")
        assert cancel_output["cancelled"] == state_unitid

        # Wait workunit detail == 'Cancelled'
        assert wait_for_workunit_state(
            node1_controller,
            state_unitid,
            expected_detail="Canceled",
            timeout_seconds=wait_for,
        ), "Workunit timed out and never finished"

        # Get work list and check for the workunit detail state
        work_list = node1_controller.simple_command("work list")
        assert work_list[state_unitid]["Detail"] == "Canceled"

        node1_controller.close()



import sys
import os
import time
import select
import fcntl
import tty
import termios
import click
import json
from functools import partial
import dateutil.parser
import importlib.metadata
from .socket_interface import ReceptorControl


class IgnoreRequiredWithHelp(click.Group):
    # allows user to call --help without needing to provide required=true parameters
    def parse_args(self, ctx, args):
        try:
            return super(IgnoreRequiredWithHelp, self).parse_args(ctx, args)
        except click.MissingParameter:
            if "--help" not in args:
                raise

            # remove the required params so that help can display
            for param in self.params:
                param.required = False
            return super(IgnoreRequiredWithHelp, self).parse_args(ctx, args)


def print_json(json_data):
    click.echo(json.dumps(json_data, indent=4, sort_keys=True))


def print_message(message="", nl=True):
    click.echo(message, nl=nl)


def print_warning(message, nl=True):
    click.echo(click.style(f"Warning: {message}", fg="magenta"), err=True, nl=nl)


def print_error(message, nl=True):
    click.echo(click.style(f"ERROR: {message}", fg="red"), err=True, nl=nl)


@click.group(cls=IgnoreRequiredWithHelp)
@click.pass_context
@click.option(
    "--socket",
    envvar="RECEPTORCTL_SOCKET",
    required=True,
    show_envvar=True,
    help="Control socket address to for the Receptor connection (The default is 'unix:' for a Unix socket, use 'tcp://' for a TCP socket)",  # noqa: E501
)
@click.option(
    "--config",
    "-c",
    default=None,
    envvar="RECEPTORCTL_CONFIG",
    required=False,
    show_envvar=True,
    help="Config filename configured for receptor",
)
@click.option(
    "--tls-client",
    "tlsclient",
    default=None,
    envvar="RECEPTORCTL_TLSCLIENT",
    required=False,
    show_envvar=True,
    help="TLS client name specified in config",
)
@click.option(
    "--rootcas",
    default=None,
    help="Root CA bundle to use instead of system trust when connecting with tls",
)
@click.option("--key", default=None, help="Client private key filename")
@click.option("--cert", default=None, help="Client certificate filename")
@click.option(
    "--insecureskipverify",
    default=False,
    help="Accept any server cert",
    show_default=True,
)
def cli(ctx, socket, config, tlsclient, rootcas, key, cert, insecureskipverify):
    ctx.obj = {
        "rc": None,
        "receptorctlVersion": importlib.metadata.version("receptorctl"),
        "receptorVersion": "Unknown",
    }
    # If we got a socket parameter we can make a ReceptorControl object
    if ctx.params.get("socket", None) is not None:
        ctx.obj["rc"] = ReceptorControl(
            socket,
            config=config,
            tlsclient=tlsclient,
            rootcas=rootcas,
            key=key,
            cert=cert,
            insecureskipverify=insecureskipverify,
        )
        # Load and stash the versions
        ctx.obj["receptorVersion"] = ctx.obj["rc"].simple_command(
            '{"command":"status","requested_fields":["Version"]}'
        )["Version"]
        # If they mismatch throw a stderr warning
        if ctx.obj["receptorVersion"] != ctx.obj["receptorctlVersion"]:
            click.echo(
                click.style(
                    "Warning: receptorctl and receptor are different versions, they may not be compatible",  # noqa E501
                    fg="magenta",
                ),
                err=True,
            )


def get_rc(ctx):
    return ctx.obj["rc"]


@cli.command(help="Show the status of the Receptor network.")
@click.pass_context
@click.option("--json", "printjson", help="Print as JSON", is_flag=True)
def status(ctx, printjson):
    rc = get_rc(ctx)
    status = rc.simple_command("status")
    if printjson:
        print_json(status)
        return
    node_id = status.pop("NodeID")
    print_message(f"Node ID: {node_id}")
    version = status.pop("Version")
    print_message(f"Version: {version}")
    sysCPU = status.pop("SystemCPUCount")
    print_message(f"System CPU Count: {sysCPU}")
    sysMemory = status.pop("SystemMemoryMiB")
    print_message(f"System Memory MiB: {sysMemory}")

    longest_node = 12

    connections = status.pop("Connections", None)
    if connections:
        for conn in connections:
            length = len(conn["NodeID"])
            if length > longest_node:
                longest_node = length

    costs = status.pop("KnownConnectionCosts", None)
    if costs:
        for node in costs:
            if len(node) > longest_node:
                longest_node = len(node)

    if connections:
        for conn in connections:
            length = len(conn["NodeID"])
            if length > longest_node:
                longest_node = length
        print_message("")
        print_message(f"{'Connection':<{longest_node}} Cost")
        for conn in connections:
            print_message(f"{conn['NodeID']:<{longest_node}} {conn['Cost']}")

    if costs:
        print_message()
        print_message(f"{'Known Node':<{longest_node}} Known Connections")
        for node in costs:
            print_message(f"{node:<{longest_node}} ", nl=False)
            for peer, cost in costs[node].items():
                print_message(f"{peer}: {cost} ", nl=False)
            print_message()

    routes = status.pop("RoutingTable", None)
    if routes:
        print_message()
        print_message(f"{'Route':<{longest_node}} Via")
        for node in routes:
            print_message(f"{node:<{longest_node}} {routes[node]}")

    ads = status.pop("Advertisements", None)
    if ads:
        print_message()
        print_message(
            f"{'Node':<{longest_node}} Service   Type       Last Seen             Tags"
        )
        for ad in ads:
            time = dateutil.parser.parse(ad["Time"])
            if ad["ConnType"] == 0:
                conn_type = "Datagram"
            elif ad["ConnType"] == 1:
                conn_type = "Stream"
            elif ad["ConnType"] == 2:
                conn_type = "StreamTLS"
            last_seen = f"{time:%Y-%m-%d %H:%M:%S}"
            print_message(
                f"{ad['NodeID']:<{longest_node}} {ad['Service']:<9} {conn_type:<10} {last_seen:<21} {'-' if (ad['Tags'] is None) else str(ad['Tags']):<16}"  # noqa: E501
            )

    def print_worktypes(header, isSecure):
        printOnce = True
        seen_nodes = []
        for ad in ads:
            commands = ad["WorkCommands"]
            if not commands:
                continue
            workTypes = []
            for c in commands:
                wT = c["WorkType"]
                if c["Secure"] == isSecure:
                    workTypes.append(wT)
            if not workTypes:
                continue
            node = ad["NodeID"]
            if node in seen_nodes:
                continue
            else:
                seen_nodes.append(node)
            workTypes = ", ".join(workTypes)
            if printOnce:
                print_message()
                print_message(f"{'Node':<{longest_node}} {header}")
                printOnce = False
            print_message(f"{ad['NodeID']:<{longest_node}} ", nl=False)
            print_message(workTypes)

    if ads:
        print_worktypes("Work Types", False)
        print_worktypes("Secure Work Types", True)

    if status:
        print_message("Additional data returned from Receptor:")
        print_json(status)


@cli.command(help="Ping a Receptor node.")
@click.pass_context
@click.argument("node")
@click.option("--count", default=4, help="Number of pings to send", show_default=True)
@click.option(
    "--delay", default=1.0, help="Time to wait between pings", show_default=True
)
def ping(ctx, node, count, delay):
    rc = get_rc(ctx)
    ping_error = False
    for i in range(count):
        results = rc.simple_command(f"ping {node}")
        if "Success" in results and results["Success"]:
            print_message(f"Reply from {results['From']} in {results['TimeStr']}")
        else:
            ping_error = True
            if "From" in results and "TimeStr" in results:
                print_error(
                    f"{results['Error']} from {results['From']} in {results['TimeStr']}"
                )
            else:
                print_error(f"{results['Error']}")
        if i < count - 1:
            time.sleep(delay)
    if ping_error:
        sys.exit(2)


@cli.command(help="Reload receptor configuration.")
@click.pass_context
def reload(ctx):
    rc = get_rc(ctx)
    results = rc.simple_command("reload")
    if "Success" in results and results["Success"]:
        print_message("Reload successful")
    else:
        print_error(f"{results['Error']}")
        if "ERRORCODE 3" in results["Error"]:
            sys.exit(3)
        elif "ERRORCODE 4" in results["Error"]:
            sys.exit(4)
        else:
            sys.exit(5)


@cli.command(help="Do a traceroute to a Receptor node.")
@click.pass_context
@click.argument("node")
def traceroute(ctx, node):
    rc = get_rc(ctx)
    results = rc.simple_command(f"traceroute {node}")
    for resno in sorted(results, key=lambda r: int(r)):
        resval = results[resno]
        if "Error" in resval:
            print_error(
                f"{resno}: Error {resval['Error']} from {resval['From']} in {resval['TimeStr']}"
            )
        else:
            print_message(f"{resno}: {resval['From']} in {resval['TimeStr']}")


@cli.command(help="Connect the local terminal to a Receptor service on a remote node.")
@click.pass_context
@click.argument("node")
@click.argument("service")
@click.option(
    "--raw", "-r", default=False, is_flag=True, help="Set terminal to raw mode"
)
@click.option(
    "--tls-client",
    "tlsclient",
    type=str,
    default="",
    help="TLS client config name used when connecting to remote node",
)
def connect(ctx, node, service, raw, tlsclient):
    rc = get_rc(ctx)
    rc.connect_to_service(node, service, tlsclient)

    stdin_tattrs = termios.tcgetattr(sys.stdin)
    stdin_fcntl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
    fcntl.fcntl(sys.stdin, fcntl.F_SETFL, stdin_fcntl | os.O_NONBLOCK)
    if raw and sys.stdin.isatty():
        tty.setraw(sys.stdin.fileno(), termios.TCSAFLUSH)
        new_term = termios.tcgetattr(sys.stdin)
        new_term[3] = new_term[3] & ~termios.ISIG
        termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, new_term)

    try:
        while True:
            r, _, _ = select.select([rc._socket, sys.stdin], [], [])
            for readable in r:
                if readable is rc._socket:
                    data = rc._socket.recv(4096)
                    if not data:
                        return
                    sys.stdout.write(data.decode())
                    sys.stdout.flush()
                else:
                    data = sys.stdin.read()
                    if not data:
                        return
                    rc._socket.send(data.encode())
    finally:
        termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, stdin_tattrs)
        print_message()


@cli.group(help="Commands related to unit-of-work processing")
def work():
    pass


@cli.command(help="Show version information for receptorctl and the receptor node")
@click.pass_context
def version(ctx):
    print_message(f"receptorctl  {ctx.obj['receptorctlVersion']}")
    print_message(f"receptor     {ctx.obj['receptorVersion']}")


@work.command(name="list", help="List known units of work.")
@click.option("--quiet", "-q", is_flag=True, help="Only list unit IDs with no detail")
@click.option(
    "--node",
    default=None,
    type=str,
    help="Receptor node to list work from. Defaults to the local node.",
)
@click.option(
    "--unit_id",
    type=str,
    required=False,
    default="",
    help="Only show detail for a specific unit id",
)
@click.option(
    "--tls-client",
    "tlsclient",
    type=str,
    default="",
    help="TLS client config name used when connecting to remote node",
)
@click.pass_context
def list_units(ctx, unit_id, node, tlsclient, quiet):
    rc = get_rc(ctx)
    if node:
        rc.connect_to_service(node, "control", tlsclient)
        rc.handshake()
    if unit_id:
        unit_id = " " + unit_id
    work = rc.simple_command("work list" + unit_id)
    if quiet:
        for k in work.keys():
            print_message(k)
    else:
        print_json(work)


@work.command(help="Submit a new unit of work.")
@click.pass_context
@click.argument("worktype", type=str, required=True)
@click.option(
    "--node",
    type=str,
    help="Receptor node to run the work on. Defaults to the local node.",
)
@click.option(
    "--payload",
    "-p",
    type=str,
    help="File containing unit of work data. Use - for stdin.",
)
@click.option(
    "--payload-literal",
    "-l",
    type=str,
    help="Use the command line string as the literal unit of work data.",
)
@click.option("--no-payload", "-n", is_flag=True, help="Send an empty payload.")
@click.option(
    "--tls-client",
    "tlsclient",
    type=str,
    default="",
    help="TLS client used when submitting work to a remote node",
)
@click.option(
    "--ttl",
    type=str,
    default="",
    help="Time to live until remote work must start, e.g. 1h20m30s or 30m10s",
)
@click.option("--signwork", help="Digitally sign remote work submissions", is_flag=True)
@click.option(
    "--follow",
    "-f",
    help="Remain attached to the job and print its results to stdout",
    is_flag=True,
)
@click.option("--rm", help="Release unit after completion", is_flag=True)
@click.option(
    "--param",
    "-a",
    help="Additional Receptor parameter (key=value format)",
    multiple=True,
)
@click.argument("cmdparams", type=str, required=False, nargs=-1)
def submit(
    ctx,
    worktype,
    node,
    payload,
    no_payload,
    payload_literal,
    tlsclient,
    ttl,
    signwork,
    follow,
    rm,
    param,
    cmdparams,
):
    pcmds = 0
    if payload:
        pcmds += 1
    if no_payload:
        pcmds += 1
    if payload_literal:
        pcmds += 1
    if pcmds < 1:
        print_error("Must provide one of --payload, --no-payload or --payload-literal.")
        sys.exit(1)
    if pcmds > 1:
        print_error(
            "Cannot provide more than one of --payload, --no-payload and --payload-literal."
        )
        sys.exit(1)
    if rm and not follow:
        print_warning("using --rm without --follow. Unit results will never be seen.")
    if payload_literal:
        payload_data = f"{payload_literal}\n".encode()
    elif no_payload:
        payload_data = "".encode()
    else:
        if payload == "-":
            payload_data = sys.stdin.buffer
        else:
            try:
                payload_data = open(payload, "rb")
            except Exception as e:
                print_error(f"Failed to load payload file: {e}")
                sys.exit(1)
    unitid = None
    try:
        params = dict(s.split("=", 1) for s in param)
        if cmdparams:
            allparams = []
            if "params" in params:
                allparams.append(params["params"])
            allparams.extend(cmdparams)
            params["params"] = " ".join(allparams)
        if node == "":
            node = None
        rc = get_rc(ctx)
        work = rc.submit_work(
            worktype,
            payload_data,
            node=node,
            tlsclient=tlsclient,
            ttl=ttl,
            signwork=signwork,
            params=params,
        )
        result = work.pop("result")
        unitid = work.pop("unitid")
        if follow:
            ctx.invoke(results, unit_id=unitid)
        else:
            print_message(f"Result: {result}")
            print_message(f"Unit ID: {unitid}")
    except Exception as e:
        print_error(e)
        sys.exit(101)
    finally:
        if rm and unitid:
            op_on_unit_ids(ctx, "release", [unitid])


@work.command(help="Get results for a previously or currently running unit of work.")
@click.pass_context
@click.argument("unit_id", type=str, required=True)
def results(ctx, unit_id):
    rc = get_rc(ctx)
    resultsfile = rc.get_work_results(unit_id)
    for text in iter(partial(resultsfile.readline, 256), b""):
        sys.stdout.buffer.write(text)
        sys.stdout.buffer.flush()
    rc = get_rc(ctx)
    status = rc.simple_command(f"work status {unit_id}")
    state = status.pop("State", 0)
    if state == 3:  # Failed
        detail = status.pop("Detail", "Unknown")
        print_error(f"Remote unit failed: {detail}\n")
        sys.exit(1)


def op_on_unit_ids(ctx, op, unit_ids):
    rc = get_rc(ctx)
    for unit_id in unit_ids:
        try:
            res = list(rc.simple_command(f"work {op} {unit_id}").items())[0]
            print_message(f"({res[1]}, {res[0]})")
        except Exception as e:
            print_error(f"{unit_id}: ERROR: {e}")
            sys.exit(1)


@work.command(help="Cancel (kill) one or more units of work.")
@click.argument("unit_ids", nargs=-1)
@click.pass_context
def cancel(ctx, unit_ids):
    if len(unit_ids) == 0:
        print_warning("No unit IDs supplied: Not doing anything")
        return
    print_message("Cancelled:")
    op_on_unit_ids(ctx, "cancel", unit_ids)


@work.command(help="Release (delete) one or more units of work.")
@click.option(
    "--force",
    help="Delete locally even if we can't reach the remote node",
    is_flag=True,
)
@click.option("--all", help="Delete all work units", is_flag=True)
@click.argument("unit_ids", nargs=-1)
@click.pass_context
def release(ctx, force, all, unit_ids):
    if len(unit_ids) == 0 and not all:
        print_warning("No unit IDs supplied: Not doing anything")
        return
    op = "release" if not force else "force-release"
    print_message("Released:")
    if all:
        rc = get_rc(ctx)
        work = rc.simple_command("work list")
        op_on_unit_ids(ctx, op, work.keys())
    else:
        op_on_unit_ids(ctx, op, unit_ids)


def run():
    try:
        cli.main(sys.argv[1:], standalone_mode=False)
    except click.exceptions.Abort:
        pass
    except Exception as e:
        print_error(e)
        sys.exit(1)
    sys.exit(0)
